;; Generated package description from soap-client.el  -*- no-byte-compile: t -*-
(define-package "soap-client" "3.1.5" "Access SOAP web services" '((cl-lib "0.6.1")) :commit "84efc93a5525f85659955b113c427a27c80c2a71" :authors '(("Alexandru Harsanyi" . "AlexHarsanyi@gmail.com")) :maintainer '("Alexandru Harsanyi" . "AlexHarsanyi@gmail.com") :keywords '("soap" "web-services" "comm" "hypermedia") :url "https://github.com/alex-hhh/emacs-soap-client")
